VERSION = '0.0.6'
VERSION_HASH = ''
BUILD_DATE = '2023-11-19 11:57:03 UTC'

HEADER_REQUEST_ID = "X-CONVERGENCE-REQUEST-ID"
HEADER_CALLER_SERVICE = "X-CONVERGENCE-CALLER-SERVICE"
HEADER_CALLER_SERVICE_HASH = "X-CONVERGENCE-CALLER-SERVICE-HASH"
HEADER_CALLER_SERVICE_VERSION = "X-CONVERGENCE-CALLER-SERVICE-VERSION"
HEADER_PARENT_REQUEST_ID = "X-CONVERGENCE-PARENT-REQUEST-ID"
